// license: AGPLv3 or later
// copyright 2023 zamfofex

#include <unistd.h>
#include <stdlib.h>
#include <termios.h>
#include <time.h>
#include <fcntl.h>
#include <stdint.h>
#include <string.h>
#include <sys/mman.h>
#include <sys/ioctl.h>
#include <linux/fb.h>
#include <linux/kd.h>
#include <linux/input.h>

#include <mimimi/engines/fbdev.h>
#include <mimimi/assets.h>
#include <mimimi/geometry.h>
#include <mimimi/engines.h>
#include <mimimi/behaviors.h>
#include <mimimi/malloc.h>
#include <mimimi/chapters.h>

static struct mimimi_size mimimi_fbdev_size = {512, 256};
static struct termios mimimi_fbdev_tio;

static void mimimi_fbdev_exit(void)
{
	ioctl(0, KDSETMODE, KD_TEXT);
	tcsetattr(0, TCSANOW, &mimimi_fbdev_tio);
}

struct mimimi_fbdev
{
	unsigned char *buffer;
	int width;
	int height;
	int x_offset;
	int y_offset;
	int line_length;
	int bytes_per_pixel;
};

static void mimimi_fbdev_stamp(void *data, int x0, int y0, void *texture)
{
	struct mimimi_image *image = texture;
	struct mimimi_fbdev *fbdev_data = data;
	
	static unsigned char rch[] = {0x11, 0x44, 0x77, 0x99, 0xCC, 0xFF};
	static unsigned char gch[] = {0x11, 0x33, 0x55, 0x77, 0x99, 0xBB, 0xDD, 0xFF};
	static unsigned char bch[] = {0x22, 0x55, 0x88, 0xBB, 0xEE};
	
	for (int y = 0 ; y < image->height ; y++)
	for (int x = 0 ; x < image->width ; x++)
	{
		if (y + y0 < 0) continue;
		if (x + x0 < 0) continue;
		
		if (y + y0 >= fbdev_data->height) continue;
		if (x + x0 >= fbdev_data->width) continue;
		
		int offset_x = (x + x0 + fbdev_data->x_offset) * fbdev_data->bytes_per_pixel;
		int offset_y = (y + y0 + fbdev_data->y_offset) * fbdev_data->line_length;
		
		int offset = offset_x + offset_y;
		
		unsigned char color = image->colors[x + y * image->width];
		if (color == 0) continue;
		
		if (fbdev_data->bytes_per_pixel == 4)
			fbdev_data->buffer[offset + 3] = 0xFF;
		
		if (color < 0x10)
		{
			unsigned char ch = color * 0x11;
			fbdev_data->buffer[offset + 2] = ch;
			fbdev_data->buffer[offset + 1] = ch;
			fbdev_data->buffer[offset + 0] = ch;
			continue;
		}
		
		color -= 0x10;
		unsigned char r, g, b;
		r = rch[(color / 40) % 6];
		g = gch[(color / 5) % 8];
		b = bch[(color / 1) % 5];
		
		fbdev_data->buffer[offset + 2] = r;
		fbdev_data->buffer[offset + 1] = g;
		fbdev_data->buffer[offset + 0] = b;
	}
}

static void *mimimi_fbdev_texture(void *data, struct mimimi_image *image)
{
	(void) data;
	return image;
}

static void mimimi_fbdev_invalidate(void *data, void *texture)
{
	(void) data;
	(void) texture;
}

void mimimi_fbdev(struct mimimi_chapter *(*start)(struct mimimi_engine *engine))
{
	int fd = open("/dev/fb0", O_RDWR);
	if (fd < 0) return;
	
	struct fb_fix_screeninfo finfo;
	if (ioctl(fd, FBIOGET_FSCREENINFO, &finfo) < 0) return;
	
	struct fb_var_screeninfo vinfo;
	if (ioctl(fd, FBIOGET_VSCREENINFO, &vinfo) < 0) return;
	
	void *aux = malloc(finfo.smem_len);
	if (aux == NULL) return;
	
	unsigned char *buffer = mmap(NULL, finfo.smem_len, PROT_WRITE, MAP_SHARED, fd, 0);
	if ((intptr_t) buffer == -1) return;
	
	int bytes_per_pixel = vinfo.bits_per_pixel / 8;
	
	if (bytes_per_pixel != 4 && bytes_per_pixel != 3) return;
	
	unsigned int width = mimimi_fbdev_size.width;
	unsigned int height = mimimi_fbdev_size.height;
	
	if (vinfo.xres < width)
		width = vinfo.xres;
	if (vinfo.yres < height)
		height = vinfo.yres;
	
	tcgetattr(0, &mimimi_fbdev_tio);
	tcflag_t flag = mimimi_fbdev_tio.c_lflag;
	mimimi_fbdev_tio.c_lflag &= ~ICANON & ~ECHO;
	
	atexit(&mimimi_fbdev_exit);
	
	// ioctl(0, KDSETMODE, KD_GRAPHICS);
	tcsetattr(0, TCSANOW, &mimimi_fbdev_tio);
	mimimi_fbdev_tio.c_lflag = flag;
	
	struct mimimi_fbdev data = {aux, width, height, vinfo.xoffset, vinfo.yoffset, finfo.line_length, bytes_per_pixel};
	
	struct mimimi_engine engine;
	engine.data = &data;
	engine.stamp = &mimimi_fbdev_stamp;
	engine.texture = &mimimi_fbdev_texture;
	engine.invalidate = &mimimi_fbdev_invalidate;
	engine.size = &mimimi_fbdev_size;
	engine.allocator = mimimi_malloc;
	struct mimimi_chapter *chapter = (*start)(&engine);
	struct mimimi_behavior *behavior = chapter->behavior;
	
	static char *file_names[] =
	{
		"/dev/input/event0", "/dev/input/event1", "/dev/input/event2", "/dev/input/event3",
		"/dev/input/event4", "/dev/input/event5", "/dev/input/event6", "/dev/input/event7",
		"/dev/input/event8", "/dev/input/event9", "/dev/input/event10", "/dev/input/event11",
		"/dev/input/event12", "/dev/input/event13", "/dev/input/event14", "/dev/input/event15",
		"/dev/input/event16", "/dev/input/event17", "/dev/input/event18", "/dev/input/event19",
		"/dev/input/event20", "/dev/input/event21", "/dev/input/event22", "/dev/input/event23",
		"/dev/input/event24", "/dev/input/event25", "/dev/input/event26", "/dev/input/event27",
		"/dev/input/event28", "/dev/input/event29", "/dev/input/event30", "/dev/input/event31",
	};
	
	int in_count;
	int in[32];
	for (in_count = 0 ; in_count < 32 ; in_count++)
	{
		int fd = open(file_names[in_count], O_RDONLY);
		if (fd < 0) break;
		in[in_count] = fd;
	}
	
	struct timespec past;
	
	for (;;)
	{
		struct timespec tv;
		clock_gettime(CLOCK_MONOTONIC, &tv);
		
		long int delay =
			(tv.tv_sec - past.tv_sec) * 1000000000 +
			(tv.tv_nsec - past.tv_nsec);
		
		if (delay < 0) delay = 0;
		
		past.tv_sec = tv.tv_sec;
		past.tv_nsec = tv.tv_nsec;
		
		unsigned long int nsec = 0;
		if (delay < 33333333) nsec = 33333333 - delay;
		
		tv.tv_sec = nsec / 1000000000;
		tv.tv_nsec = nsec % 1000000000;
		
		int nfds = 0;
		fd_set fds;
		FD_ZERO(&fds);
		for (int i = 0 ; i < in_count ; i++)
		{
			if (nfds < in[i]) nfds = in[i];
			FD_SET(in[i], &fds);
		}
		
		for (;;)
		{
			int available = pselect(nfds, &fds, NULL, NULL, &tv, NULL);
			if (available == -1) return;
			if (available == 0) break;
			
			tv.tv_sec = 0;
			tv.tv_nsec = 0;
			
			for (int i = 0 ; i < in_count ; i++)
			{
				int fd = in[i];
				if (FD_ISSET(fd, &fds) == 0) continue;
				
				struct input_event event;
				int size = sizeof event;
				ssize_t r = read(fd, &event, size);
				if (r < size) return;
				
				if (event.type != EV_KEY) continue;
				if (event.value == 2) continue;
				
				switch (event.code)
				{
				case KEY_LEFT:
				case KEY_A:
					chapter->left = event.value;
					break;
				case KEY_RIGHT:
				case KEY_D:
					chapter->right = event.value;
					break;
				case KEY_Q:
					if (event.value == 0)
						return;
					break;
				}
			}
		}
		
		(*behavior->behave)(behavior->data);
		memcpy(buffer, aux, finfo.smem_len);
	}
}
